'''
google/utils.py: general storage utility functions

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


from googleapiclient.errors import HttpError
from googleapiclient import (
    http,
    discovery
)

from oauth2client.client import GoogleCredentials
from oauth2client.service_account import ServiceAccountCredentials

from som.logger import bot

from som.utils import (
    get_installdir,
    read_file,
    run_command
)

from subprocess import (
    Popen,
    PIPE,
    STDOUT
)

import tempfile
import zipfile

from glob import glob
import httplib2
import inspect
import imp
import os
import json
import pickle
import re
import requests
from retrying import retry
import shutil
import simplejson
import sys
import tempfile
import time
import zipfile


######################################################################################
# Testing/Retry Functions
######################################################################################

def stop_if_result_none(result):
    '''stop if result none will return True if we should not retry 
    when result is none, False otherwise using retrying python package
    '''
    do_retry = result is not None
    return do_retry


# Simple default retrying for calls to api
doretry = retry(wait_exponential_multiplier=1000,wait_exponential_max=10000,stop_max_attempt_number=10)



##########################################################################################
# GOOGLE GENERAL API #####################################################################
##########################################################################################

def get_google_service(service_type=None,version=None):
    '''
    get_url will use the requests library to get a url
    :param service_type: the service to get (default is storage)
    :param version: version to use (default is v1)
    '''
    if service_type == None:
        service_type = "storage"
    if version == None:
        version = "v1"

    credentials = GoogleCredentials.get_application_default()
    return discovery.build(service_type, version, credentials=credentials) 


##########################################################################################
# GOOGLE STORAGE API #####################################################################
##########################################################################################
    
def get_bucket(storage_service,bucket_name):
    req = storage_service.buckets().get(bucket=bucket_name)
    return req.execute()


def delete_object(storage_service,bucket_name,object_name):
    '''delete_file will delete a file from a bucket
    :param storage_service: the service obtained with get_storage_service
    :param bucket_name: the name of the bucket (eg singularity-hub)
    :param object_name: the "name" parameter of the object.
    '''
    try:
        operation = storage_service.objects().delete(bucket=bucket_name,
                                                     object=object_name).execute()
    except HttpError as e:
        pass
        operation = e
    return operation


def upload_file(storage_service,bucket,bucket_path,file_path,verbose=True):
    '''get_folder will return the folder with folder_name, and if create=True,
    will create it if not found. If folder is found or created, the metadata is
    returned, otherwise None is returned
    :param storage_service: the drive_service created from get_storage_service
    :param bucket: the bucket object from get_bucket
    :param file_path: the path to the file to upload
    :param bucket_path: the path to upload to
    '''
    if bucket_path[-1] != '/':
        bucket_path = "%s/" %(bucket_path)
    bucket_path = "%s%s" %(bucket_path,os.path.basename(file_path))
    body = {'name': bucket_path }

    # Create media object with correct mimetype
    mimetype = sniff_extension(file_path,verbose=verbose)
    media = http.MediaFileUpload(file_path,
                                 mimetype=mimetype,
                                 resumable=True)
    request = storage_service.objects().insert(bucket=bucket['id'], 
                                               body=body,
                                               predefinedAcl="publicRead",
                                               media_body=media)
    return request.execute()



def list_bucket(bucket,storage_service):
    # Create a request to objects.list to retrieve a list of objects.        
    request = storage_service.objects().list(bucket=bucket['id'], 
                                             fields='nextPageToken,items(name,size,contentType)')
    # Go through the request and look for the folder
    objects = []
    while request:
        response = request.execute()
        objects = objects + response['items']
    return objects


def get_image_path(repo_url,commit):
    '''get_image_path will determine an image path based on a repo url, removing
    any token, and taking into account urls that end with .git.
    :param repo_url: the repo url to parse:
    :param commit: the commit to use
    '''
    repo_url = repo_url.split('@')[-1].strip()
    if repo_url.endswith('.git'):
        repo_url =  repo_url[:-4]
    return "%s/%s" %(re.sub('^http.+//www[.]','',repo_url),commit)




#####################################################################################
# METADATA
#####################################################################################

def get_storage_fields(obj):
    '''get_storage_fields will return a subset of storage object
    fields, to be saved as strings with a datastore object.
    '''                                   
    fields = {'storage_bucket':obj['bucket'],
              'storage_md5Hash':obj['md5Hash'],
              'storage_updated':obj['updated'],
              'storage_download':obj['mediaLink'],
              'storage_metadataLink':obj['selfLink'],
              'storage_crc32c':obj['crc32c'],
              'storage_etag':obj['etag'],
              'storage_owner':obj['owner']['entity'],
              'storage_name':obj['name'],
              'storage_id':obj['id'],
              'storage_contentType':obj['contentType'],
              'storage_size':obj['size']}
    return fields


def get_metadata(key):
    '''get_metadata will return metadata about an instance from within it.
    :param key: the key to look up
    '''
    headers = {"Metadata-Flavor":"Google"}
    url = "http://metadata.google.internal/computeMetadata/v1/instance/attributes/%s" %(key)        
    response = api_get(url=url,headers=headers)

    # Successful query returns the result
    if response.status_code == 200:
        return response.text
    else:
        bot.error("Error retrieving metadata %s, returned response %s" %(key,
                                                                         response.status_code))
    return None



######################################################################################
# Extensions and Files
######################################################################################


def sniff_extension(file_path,verbose=True):
    '''sniff_extension will attempt to determine the file type based on the extension,
    and return the proper mimetype
    :param file_path: the full path to the file to sniff
    :param verbose: print stuff out
    '''
    mime_types =    { "xls": 'application/vnd.ms-excel',
                      "xlsx": 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
                      "xml": 'text/xml',
                      "ods": 'application/vnd.oasis.opendocument.spreadsheet',
                      "csv": 'text/plain',
                      "tmpl": 'text/plain',
                      "pdf":  'application/pdf',
                      "php": 'application/x-httpd-php',
                      "jpg": 'image/jpeg',
                      "png": 'image/png',
                      "gif": 'image/gif',
                      "bmp": 'image/bmp',
                      "txt": 'text/plain',
                      "doc": 'application/msword',
                      "js": 'text/js',
                      "swf": 'application/x-shockwave-flash',
                      "mp3": 'audio/mpeg',
                      "zip": 'application/zip',
                      "rar": 'application/rar',
                      "tar": 'application/tar',
                      "arj": 'application/arj',
                      "cab": 'application/cab',
                      "html": 'text/html',
                      "htm": 'text/html',
                      "default": 'application/octet-stream',
                      "folder": 'application/vnd.google-apps.folder',
                      "img" : "application/octet-stream" }

    ext = os.path.basename(file_path).split('.')[-1]

    mime_type = mime_types.get(ext,None)

    if mime_type == None:
        mime_type = mime_types['txt']

    if verbose==True:
        bot.info("%s --> %s" %(file_path,mime_type))

    return mime_type



'''
google/modals.py: models for datastore

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from som.api.google.storage.validators import (
    validate_model
)

import google.cloud.datastore as datastore
from som.api.google.storage.datastore import (
    get_key_filters,
    parse_keys
)

from som.logger import bot
import datetime
import collections
import sys
import os

######################################################################################
# Object Manager
######################################################################################

class BatchManager:
    '''a batch manager is bucket to hold multiple objects to filter, query, etc.
    It a way to compile a set, and then run through a transaction. It must either
    be instantiated with a client, or one is generated dynamically.
    '''

    def __init__(self,client=None):
        if client is None:
            client = datastore.Client()
        self.client = client
        self.tasks = []
        self.queries = []

    def get_kinds(self):
        query = self.client.query(kind='__kind__')
        query.keys_only()
        return [entity.key.id_or_name for entity in query.fetch()]


    def get(self,kind,keys=None,limit=None,field=None,keys_only=False,ancestor=None):
        if keys is None:
            return self.query(kind=kind,
                              limit=limit,
                              ancestor=ancestor,
                              keys_only=keys_only)
        else:
            if not isinstance(keys,list):
                keys = [keys]
            if field is None:
                return self.client.get_multi(keys)
            else:
                for key in keys:
                    query = self.client.query(kind=kind.capitalize())
                    query.add_filter(field,'=',key)
                    self.add(query)
                return self.runQueries()                    


    def add(self,task):
        '''return all tasks in the set
        '''
        if isinstance(task,datastore.query.Query):
            self.queries.append(task)
        else:
            if not isinstance(task,datastore.Entity):
                task = task._Entity
            self.tasks.append(task)


    def delete(self,keys):
        '''delete a set of model objects based on user provided keys'''
        with parse_keys(keys) as keys:
            return self.client.delete_multi(keys) # note sure if returning this 
                                                  # returns a status or nothing
                                                  # same for get, need to test
                                                  # (currently on airplane)



    def runInsert(self,clear_queue=True):
       '''runInsert will run a transaction for a set of tasks
       '''
       tasks = None
       if len(self.tasks) > 0:
           tasks = self.tasks
           with self.client.transaction():
               self.client.put_multi(tasks)
           if clear_queue:
               self.tasks = []
       return tasks
       

    def runQueries(self,clear_queue=True):
       results = None
       if len(self.queries) > 0:
           results = []
           for query in self.queries:
               result = [x for x in query.fetch()]
               results = results + result
           if clear_queue:
               self.queries = []
       return results


    def query(self,kind=None,filters=None,order=None,projections=None,
              run=True,limit=None,keys_only=False,distinct_on=None,query=None,
              ancestor=None):
        '''query will run a query for some kind of model (eg 'Collection')
        :param kind: the kind of model to query
        :param filters: a list of lists of tuples, each length 3, with

             FIELD OPERATOR VALUE
             "description" "=' "this is a description" 
 
        :param order: optional one or more orderings -descending / ascending
        :param projections: if defined, return a dictionary of lists of each
        :param run: Default true to execute query. False returns query object
        :param limit: an integer limit of results to return
        :param distinct_on: one or more fields to make distinct
        :param query: if provided, will not start with basic query
        :param ancestor: if provided, query for the provided ancestor
                                
             ANCESTOR EXAMPLE
             ancestor = client.key('TaskList', 'default')

        '''
        if ancestor is not None:
            if isinstance(ancestor,list) or isinstance(ancestor,str):
                ancestor = self.client.key(*ancestor)

        if query == None:
            if kind is None:
                bot.error("You must define 'kind' to run a query.")
                sys.exit(1)
            query = self.client.query(kind=kind.capitalize(),ancestor=ancestor) # uppercase first letter

        if distinct_on is not None:
            if not isinstance(distinct_on,list):
                distinct_on = [distinct_on]
            query.distinct_on = distinct_on

        # Add filters to the query
        if filters is not None:
            for f in filters:
                query.add_filter(f[0], f[1],f[2])

        # Apply order to the query (eg, ['-priority']
        if order is not None:
            if not isinstance(order,list):
                order = [order]
            order = [x.lower() for x in order]
            query.order = order

        # Return a projection, an extraction of fields from a set
        if projections is not None:
            return self.projection(query,projections)

        if keys_only is True:
            query.keys_only()
  
        # Do we not want to run the query?
        if not run:
            return query

        result = None

        try:
            result = [x for x in query.fetch(limit=limit)]

        except (google.cloud.exceptions.BadRequest,
                google.cloud.exceptions.GrpcRendezvous):
            bot.error("Error with query.")         
            pass

        return result


    def query_key(self,kind,keys,**kwargs):
        '''query_key is an entry to query that adds a key filter to the query
        '''
        if operator == None:
            operator = '>'

        if operator not in get_key_filters():
            bot.error("%s is not a valid operator." %operator)
            sys.exit(1)

        keys = parse_keys(keys)
        query = self.client.query(kind=kind)
        for key in keys:
            query.key_filter(key, operator)

        return self.query(**kwargs)



    def query_date(self,kind,startyear=None,startday=None,startmonth=None,
                   endyear=None,endday=None,endmonth=None,**kwargs):
        '''query_date is an entry to query that sets up a date filter.
        '''
        start_date = None
        end_date=None
        if startyear is not None and startday is not None and startmonth is not None:
            start_date = datetime.datetime(startyear, startmonth, startday)
        if endyear is not None and endmonth is not None and endday is not None:
            end_date = datetime.datetime(endyear, endmonth, endday)
        if start_date == None and end_date == None:
            bot.warning('''Both start date and end date are null, did you provide
                           year, month, day for one or both?''')
     
        query = self.client.query(kind=kind)
        if start_date is not None:
            query.add_filter('created', '>', start_date)
        if end_date is not None:
            query.add_filter('created', '<', end_date)
     
        return self.query(**kwargs)
        

    def projection(self,query,projections,add_loners=False):
        '''extract a subset of a query, meaning a dictionary with key value
        pairs for field/results lists. By default, to preserve indexing,
        only objects with all projection fields are added to the result.
        You can change this behavior by setting add_loners = True
        ''' 
        if not isinstance(projections,list):
            projections = [projections]

        # Find a more elegant way to do this
        results = dict()
        for projection in projections:
            results[projection] = []
                    
        for task in query.fetch():
            loner = False
            for p in projections: 
                if p not in task:
                    longer=True
                    break

            # Find more elegant way to not need to loop twice    
            for p in projections:
                if not loner or (loner and add_loners):
                    results[p].append(task[p])

        return results




######################################################################################
# Base Client Class
######################################################################################

# https://cloud.google.com/datastore/docs/concepts/entities#datastore-basic-entity-python

class ModelBase:

    def __init__(self,client=None):
        '''A ModelBase is a controller for a general DataStore
        Entity. Most of the variables in init are for initial
        validation, and further fields etc are stored with
        the Entity itself (_Entity)
        :param client: the datastore client, should be passed on init
        :param _fields: intiial fields to validate on model creation
        :param _key: the initial key used for the entity
        :param _Entity: the final generated (validated) entity
        '''
        if client is None:
            client = datastore.Client()
        self.client = client
        self._fields = validate_model(self.model['fields'])
        self._key = self.client.key(*self.model['key'])
        self._Entity = None
        if "exclude_from_indexes" in self.model:
            self._exclude_from_indexes = self.model['exclude_from_indexes'] # Can be None
        else:
            self._exclude_from_indexes = None

    def __repr__(self):
        if self._Entity is None:
            return "Entity:None"
        return str(self._Entity.key)

    def __str__(self):
        if self._Entity is None:
            return "Entity:None"
        return str(self._Entity.key)


    def get_keypath(self):
        return list(self._Entity.key.flat_path)
      
    def get_name(self):
        return list(self._Entity.key.flat_path)[-1]


    def update_fields(self,new_fields,add_new=True):
        '''update fields will update the model's fields with an input dictionary
        new fields are added if add_new=True (default). This does not by default save
        the entity.
        '''
        for key,value in new_fields.items():
            if key not in self._Entity.keys():
                bot.debug("adding %s to Entity" %(key))
                self._Entity[key] = value
            else:
                if add_new == True:
                    bot.debug("%s found existing in Entity, overwriting" %(key))
                    self._Entity[key] = value


    def get_or_create(self):
        '''get_or_create will get or create an object using a transaction.
        If an entity object already exists with the object, it is overridden
        '''
        with self.client.transaction():
           self._Entity = self.client.get(*self._key)
           if not self._Entity:
               self._Entity = self.create()
        return self._Entity


    def setup(self,fields=None):
        '''setup will generate a new Entity without saving it (for batches)
        '''
        if self._exclude_from_indexes is not None:
            self._Entity = datastore.Entity(key=self._key,
                                            exclude_from_indexes=self._exclude_from_indexes)
        else:
            self._Entity = datastore.Entity(self._key)
        if fields is not None:
            self.update_fields(fields)


    def save(self):
        '''save just calls put for the current _Entity'''
        self.client.put(self._Entity)


    def create(self,fields=None,save=True):
        '''create a new entity
        :param fields: additional fields to add
        :param save: save the entity to datastore (default True)
        '''
        self.setup(fields)
        self._Entity['created'] = datetime.datetime.utcnow()
        self._Entity['updated'] = datetime.datetime.utcnow()
        # Add initial fields, if defined
        for field,value in self._fields.items():
            self._Entity[field] = value
        if save:
            self.save()
        return self._Entity


    def delete(self):
        '''delete an entity'''
        key = None
        if self._Entity is not None:
            key = self._Entity.key
            self.client.delete(key)
            bot.debug("Deleting %s" %(key))
        return key


    def update_or_create(self,fields=None,save=True):
        '''update or create will update or create an entity.
        '''
        entity = self.client.get(self._key)

        # The entity is being created, add timestamp
        if not entity:
            entity = self.create(fields,save)         #insert
        else: 
            entity = self.update(fields,save)        #upsert
        return self._Entity


    def update(self,fields=None,save=True):
        '''update an entity, meaning updating the local and then pushing
        (saving) to the datastore. This overrides the datastore version
        by default. To simply get the datastore version and override
        the local, use get'''
        if self._Entity is None:
            self._Entity = self.client.get(self._key)
        self._Entity['updated'] = datetime.datetime.utcnow()
        # Update any fields
        if fields is not None:
            for field,value in fields.items():
                self._Entity[field] = value
        if save:
            self.save()
        return self._Entity


    def get(self):
        '''get an entity, and override the currently set _Entity in the model
        if it is not retrieved yet. If already retrieved, does not update.
        '''
        if self._Entity is None:
            with self.client.transaction():
                self._Entity = self.client.get(*self._key)
        return self._Entity



#!/usr/bin/env python

'''
base.py: base module for working with som api

'''

from som.logman import bot

from som.api.validators.requests import (
    validate_identifiers,
    validate_items,
    validate_item
)

from som.api.validators.utils import (
    get_universal_source
)

import sys
import re
import os

api_base = "https://api.rit.stanford.edu/identifiers/api"
api_version = "v1"


def create_identifier(entity_id,id_source,id_timestamp=None,items=None,custom_fields=None,sources=None):
    '''create_identifier will return a json (dict) of required elements for an entity,
    a field to go under an identifier. The following is required:
    :param entity_id: mandatory key for uniquely identifying the entity (e.g. "1234567-8")
    :param id_source: mandatory issuer for the above id (e.g., "stanford")
    :param id_timestamp: when the id was looked up, to help with changed/merged ids (optional)
    :param custom_fields: not required, a dictionary of custom key pairs to include as data
    :param sources: a custom list of sources (eg, ['stanford']) to accept. If not defined, defaults
    are used
    :returns entity: an entity data structure, not wrapped, or None if not valid
    '''
    entity = {"id":entity_id,
              "id_source":id_source}

    # Custom fields and id_timestamp are optional
    if custom_fields != None:
        entity['custom_fields'] = custom_fields 

    if id_timestamp != None:
        entity["id_timestamp"] = id_timestamp

    if items != None:
        if not isinstance(items,list):
            items = [items]
        entity['items'] = items

    valid = validate_identifiers(identifiers=entity,
                                 sources=sources)
    if valid == True:
        return entity

    return None


def create_items(item_ids,id_sources,sources=None,custom_fields=None,verbose=False):
    '''create items is a wrapper for create items, taking in lists of item_id and id_source.
    if id_source is not a list, it's assumed to be equivalent for all items.
    :param item_ids: should be a list of items.
    :param id_sources: if a list is given, it must be equal in length to item ids. Otherwise,
    the common identifier (one term) is used for all items.
    :param custom_fields: not required, a list of dict of custom key pairs to include as data, or a single
    custom thing to use as data.
    :param sources: a custom list of sources (eg, ['stanford']) to accept. If not defined, defaults
    are used
    :param verbose: verbose is set to False, assuming we have multiple items! This will simply
    not print successful validations - errors are still printed for user.
    '''
    items = []

    # If item_ids is a string, assume one and put into list
    if not isinstance(item_ids,list):
        item_ids = [item_ids]    

    # If id_source is a list, must be equal in length to ids
    universal_source = get_universal_source(source=id_sources,
                                            comparator=item_ids)

    # If custom_fields is a list, must be equal in length to ids
    universal_field = get_universal_source(source=custom_fields,
                                           comparator=item_ids)

    # Iterate through items to return list of dict
    for i in range(len(item_ids)):

        # Do we have a shared item source?
        if universal_source != None:
            item_source = universal_source
        else:
            item_source = id_sources[i]

        # Do we have a shared custom_fields?
        if universal_field != None:
            item_field = universal_field
        else:
            item_field = custom_fields[i]

        item = create_item(item_id=item_ids[i],
                           id_source=item_source,
                           custom_fields=item_field,
                           sources=sources,
                           verbose=verbose)
    
        items.append(item)
    return items


def create_item(item_id,id_source,custom_fields,sources=None,verbose=True):
    '''create item will take input for an item, and return the item if it's valid
    :param item_id: the item id to store, mandatory
    :param id_source: mandatory issuer for the above id (e.g., "pacs")
    :param custom_fields: not required, a dictionary of custom key pairs to include as data
    :param id_timestamp: when the id was looked up, to help with changed/merged ids (optional)
    :param sources: a custom list of sources (eg, ['stanford']) to accept. If not defined, defaults
    '''
    item = {"id":item_id,
            "id_source":id_source}

    # Custom fields and id_timestamp are optional
    if custom_fields != None:
        item['custom_fields'] = custom_fields 

    if id_timestamp != None:
        item["id_timestamp"] = id_timestamp

    valid = validate_item(item,sources=sources,
                          custom_fields=custom_fields,
                          verbose=verbose)
    if valid == True:
        return item

    return None



#!/usr/bin/env python

'''

cli.py: command line client for som (som-validator)

'''

from som.version import __version__
import optparse
import os
import re
import shutil
import sys
import tempfile

def get_parser():


    parser = optparse.OptionParser(description='validate folder and compressed file structures of text and images',
                                   usage="usage: %prog [options]",
                                   version=__version__)

    # Name of folder or zipped file to test. Not required
    parser.add_option("--test", 
                      dest='test', 
                      help="folder or file to run validation test on. If none supplied, will use $PWD as folder.", 
                      type=str, 
                      default=None)

    # Change level of verbosity
    parser.add_option("--quiet", 
                      dest='quiet', 
                      action="store_true",
                      help="don't print information about validation", 
                      default=False)

    # Detect zip files in folder insteaed
    parser.add_option("--detect-compressed", 
                      dest='detect_compressed', 
                      action="store_true",
                      help="look for zip/tar.gz files in present working directory, if --test not specified or empty", 
                      default=False)

    return parser


def main():
    '''main is a wrapper for the client to hand the parser to the executable functions
    This makes it possible to set up a parser in test cases
    '''
    parser = get_parser()
    
    try:
        (args,options) = parser.parse_args()
    except:
        sys.exit(0)

    # Give the args to the main executable to run
    run(args)


def run(args):

    print("--- starting som-validator ---")

    if args.quiet == True:
        os.environ['SOM_MESSAGELEVEL'] = "CRITICAL"

    from som.logger import bot
    from som.utils import detect_compressed
    from som.wordfish.base import run_validation

    # We will keep a list of testing folders/zips
    folders = []

    # What are we testing?
    if args.test not in [None,""]:

        # specify using compressed in folder?
        if args.detect_compressed == True:
            inputs = detect_compressed(folder=args.test)
            folders = folders + inputs

        # just use folder, present working directory
        else:
            folders = [args.test]

        # run the validation tests
        run_validation(inputs=folders) # test_dir = None
                                       # clean_up = True
                                       # fail_exit = True
    else:
        print("You must specify a folder or compressed object with --test.")

if __name__ == '__main__':
    main()



#!/usr/bin/env python

'''

structures.py: generate json (dict) structures to represent folders, etc.

'''

from som.logger import bot
import tempfile
import shutil
import os
import re
import json
import sys

from som.utils import (
    untar_dir,
    unzip_dir,
    read_json
)

from validator import (
    Required, 
    Not, 
    Truthy, 
    Blank, 
    Range, 
    Equals, 
    In, 
    validate
)


def structure_dataset(dataset,testing_base=None,clean_up=True):
    '''structure_dataset is a general function to take a compressed object 
    (zip or .targz) or folder and run the correct functions to return a json
    datastructure, depending on the input types
    :param dataset: the full path to the dataset
    :param testing_base: the testing_base directory
    :param clean_up: whether to clean up the directory upon finish
    '''
    if os.path.isdir(dataset):
        structure = structure_folder(folder=dataset)
    elif re.search("[.]zip$|[.]tar[.]gz$",dataset):
        structure = structure_compressed(compressed_file=dataset,
                                         testing_base=testing_base,
                                         clean_up=clean_up)
    return structure


def structure_metadata(full_path,metadata_type=None):
    '''structure_metadata checks to see if a name (either a collection
    name, folder for an image or text) has associated metadata, indicated by
    a file of the same name (ending with json) in the parent directory of the
    named file. If no matching files are found, None is returned, and the user
    is alerted. If a matching file is found, the file is read and returned if 
    it's valid json.
    :param full_path: full path to a file or folder
    :param metadata_type: either one of collection, image, or text. Default collection
    '''
    if metadata_type == None:
        metadata_type = "collection"

    parent_dir = os.path.dirname(full_path)
    base_name = os.path.basename(full_path)
    metadata = "%s/%s.json" %(parent_dir,base_name)

    if os.path.exists(metadata):
        bot.debug('found %s metadata: %s' %(metadata_type, base_name))
        try:
            return read_json(metadata)
        except:
            bot.error('%s %s has invalid json metadata %s' %(metadata_type, base_name, metadata))
            return False
 
    else:
        bot.info('%s %s does not have metadata file %s.json' %(metadata_type, base_name, base_name))
        return None


def structure_compressed(compressed_file,testing_base=None,clean_up=False):
    '''structure_compressed will first decompress a file to a temporary location,
    and then return the file structure in the WordFish standard. 
    :param compressed_file: the file to first extract.
    :param testing_base: If not given, a temporary location will be created. Otherwise,
    a folder will be made in testing_base.
    :param clean_up: clean up (remove) extracted files/folders after test. Default False,
    so the user can access the extracted files.
    '''
    if testing_base == None:
        testing_base = tempfile.mkdtemp()
 
    dest_dir = tempfile.mkdtemp(prefix="%s/" %testing_base)
    if compressed_file.endswith('.tar.gz'):
        test_folder = untar_dir(compressed_file,dest_dir)
 
    elif compressed_file.endswith('.zip'):
        test_folder = unzip_dir(compressed_file,dest_dir)

    else:
        bot.error("Invalid compressed file type: %s, exiting." %compressed_file)
        sys.exit(1)

    # Each object in the folder (a collection)
    collection_paths = os.listdir(test_folder)
    bot.info("collections found: %s" %len(collection_paths))

    # We will return a list of structures, only of valid
    collections = []

    for col in collection_paths:
        collection_path = "%s/%s" %(test_folder,col)
        collection = structure_folder(collection_path)
        collections.append(collection)           

    if clean_up == True:
        shutil.rmtree(dest_dir)
    return collections


def structure_folder(folder,relative_path=False):
    '''structure_folder will return a json data structure to describe a collection folder.
    The collection is named according to the input data file, and so if additional metadata
    is to be added (a unique id, name, etc.) it should be done by the calling function using
    the name as a lookup.
    :param folder: the folder to generate a structure for
    :param relative_path: if True, will return relative paths (for web server)
    :returns collection: a dictionary of entity and other objects. 

    A collection is a dictionary with the following:

    { "collection":

      {
         "name": "collection1",
         "metadata" ... ,
         "entities": [ ... ]},
      }
 
    }

    A collection should be put into a collections data structure like:


     { "collections":
         
        [ ... ]

     }

    '''    

    collection = {'name': folder }
    full_path = os.path.abspath(folder)
    if relative_path == True:
        full_path = os.path.relpath(folder, os.getcwd())

    # Add any collection metadata
    metadata = structure_metadata(full_path)
    if metadata != None:
        collection['metadata'] = metadata

    # validate images, text, and metadata of the entities
    entities = structure_entities(full_path)
    if entities == None:
        bot.info("no entities found for collection %s." %(folder))              
    else:
        bot.info("adding %s valid entities to collection %s." %(len(entities),folder))              
        collection['entities'] = entities

    return {"collection" : collection}


def structure_entities(full_path):
    '''structure_entities will return a data structure with a list of
    images and text for each entity found. 
    :param full_path: the full path to the collection folder with entities

    An entity should look like the following:    
        { "entity": {
         
            "id": "12345-6",
            "images": [ ... ],
            "text": [ ... ] 
          }

        },
    '''
    entities = []
    contenders = os.listdir(full_path)
    bot.info("Found %s entity folders in collection." %len(contenders))
    if len(contenders) == 0:
        return None
     
    for contender in contenders:
        entity_path = "%s/%s" %(full_path,contender)
        entity = {'id':entity_path}

        # Does the entity have metadata?
        metadata = structure_metadata(entity_path,"entity")
        if metadata != None:
            entity['metadata'] = metadata

        entity_texts = structure_texts(entity_path)
        entity_images = structure_images(entity_path)

        # If images and text are empty for a collection, invalid
        if entity_texts == None and entity_images == None:
            bot.error("found invalid entity: does not have images or text.")
            continue
    
        # if either text or images are not valid, entities considered invalid
        if entity_texts != None:
            entity['texts'] = entity_texts
        if entity_images != None:
            entity['images'] = entity_images    
        entities.append({"entity":entity})

    return entities


def structure_texts(entity_path,acceptable_types=None):
    '''structure_texts will check an entity directory
    for a text folder. If it exists, each subfolder will return a list of
    texts and metadata
    :param entity_path the path to the top level (entity) folder
    :param acceptable_types: the valid extensions to allow
    '''    
    if acceptable_types == None:
        acceptable_types = ['txt']

    texts = structure_template(entity_path=entity_path, 
                               template_type="text",
                               acceptable_types=acceptable_types)
    return texts


def structure_images(entity_path, acceptable_types=None):
    '''structure_images will check an entity directory
    for an images folder. If it exists, each subfolder will be searched
    for images and corresponding metadata
    :param entity_path the path to the top level (entity) folder
    :param acceptable_types: the valid extensions to allow
    '''    
    if acceptable_types == None:
        acceptable_types = ['dcm','png','jpg','jpeg','nii','nii.gz']

    images = structure_template(entity_path=entity_path, 
                                template_type="images",
                                acceptable_types=acceptable_types)
    return images


def structure_template(entity_path, template_type, acceptable_types):
    '''structure_template will check an entity directory
    for an folder of a particular type, for files and metadata that
    meet a particular criteria. If needed, additional parsing
    functions can be passed to this function. 
    :param entity_path the path to the top level (entity) folder
    :param template_type: should be one of images or text
    :param acceptable_types: the valid extensions to allow
    '''
    template_path = "%s/%s" %(entity_path,template_type)
    entity_name = os.path.basename(entity_path)
    if not os.path.exists(template_path):
        bot.info("entity %s does not have %s." %(entity_name, template_type))
        return None
    
    # Let's keep track of each file
    entity_folders = os.listdir(template_path)

    for entity_folder in entity_folders:
        all_files = os.listdir(template_path)
        valids = []    # valid files, loaded or not

    # Find all valid images
    for folder in entity_folders:
        folder_path = "%s/%s" %(template_path,folder)
        all_files = os.listdir(folder_path)
        for single_file in all_files:
            parts = single_file.split('.')
            full_path = "%s/%s" %(folder_path,single_file)
            ext = '.'.join(parts[1:])
            if ext in acceptable_types:
                valid = {'original': full_path}
                metadata_file = "%s/%s.json" %(folder_path,parts[0])
                if os.path.exists(metadata_file):
                    valid['metadata'] = metadata_file
                valids.append(valid)

    # Warn the user about missing valid files, not logical given folder
    if len(valids) == 0:
        bot.warning("entity %s does not have %s." %(entity_name, template_type))
        return None
    else:
        bot.info("entity %s has %s %s" %(entity_name, len(valids), template_type))

    return valids



'''
utils.py: util functions for returning identifier objects

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


from som.logger import bot

from .validators import (
    get_universal_source,
    validate_identifiers,
    validate_items,
    validate_item
)

import sys
import re
import os


def create_lookup(response,lookup_field=None):
    '''create_identifier_lookup will take a response, which should be a list
    of results (the API returns a "results" object, and the client 
    returns the list under "results". A lookup dictionary is returned,
    indexed by "id"
    '''
    if lookup_field is None:
        lookup_field = "id"

    if not isinstance(response,list):
        response = [response]

    lookup = dict()
    for entry in response:
        key = entry[lookup_field]
        if key in lookup:
            bot.warning("%s already in lookup, will use last in list." %(key))
        lookup[key] = entry
    return lookup


def create_identifier(entity_id,id_source,id_timestamp=None,items=None,
                      custom_fields=None,sources=None):
    '''create_identifier will return a json (dict) of required elements for an entity,
    a field to go under an identifier. The following is required:
    :param entity_id: mandatory key for uniquely identifying the entity (e.g. "1234567-8")
    :param id_source: mandatory issuer for the above id (e.g., "stanford")
    :param id_timestamp: when the id was looked up, to help with changed/merged ids (optional)
    :param custom_fields: not required, a dictionary of custom key pairs to include as data
    :param sources: a custom list of sources (eg, ['stanford']) to accept. If not defined, defaults
    are used
    :returns entity: an entity data structure, not wrapped, or None if not valid
    '''
    entity = {"id":entity_id,
              "id_source":id_source}

    # Custom fields and id_timestamp are optional
    if custom_fields != None:
        entity['custom_fields'] = custom_fields 

    if id_timestamp != None:
        entity["id_timestamp"] = id_timestamp

    if items != None:
        if not isinstance(items,list):
            items = [items]
        entity['items'] = items

    valid = validate_identifiers(identifiers=entity,
                                 sources=sources)
    if valid == True:
        return entity

    return None


def create_items(item_ids,id_sources,sources=None,custom_fields=None,verbose=False):
    '''create items is a wrapper for create items, taking in lists of item_id and id_source.
    if id_source is not a list, it's assumed to be equivalent for all items.
    :param item_ids: should be a list of items.
    :param id_sources: if a list is given, it must be equal in length to item ids. Otherwise,
    the common identifier (one term) is used for all items.
    :param custom_fields: not required, a list of dict of custom key pairs to include as data, or a single
    custom thing to use as data.
    :param sources: a custom list of sources (eg, ['stanford']) to accept. If not defined, defaults
    are used
    :param verbose: verbose is set to False, assuming we have multiple items! This will simply
    not print successful validations - errors are still printed for user.
    '''
    items = []

    # If item_ids is a string, assume one and put into list
    if not isinstance(item_ids,list):
        item_ids = [item_ids]    

    # If id_source is a list, must be equal in length to ids
    universal_source = get_universal_source(source=id_sources,
                                            comparator=item_ids)

    # If custom_fields is a list, must be equal in length to ids
    universal_field = get_universal_source(source=custom_fields,
                                           comparator=item_ids)

    # Iterate through items to return list of dict
    for i in range(len(item_ids)):

        # Do we have a shared item source?
        if universal_source != None:
            item_source = universal_source
        else:
            item_source = id_sources[i]

        # Do we have a shared custom_fields?
        if universal_field != None:
            item_field = universal_field
        else:
            item_field = custom_fields[i]

        item = create_item(item_id=item_ids[i],
                           id_source=item_source,
                           custom_fields=item_field,
                           sources=sources,
                           verbose=verbose)
    
        items.append(item)
    return items


def create_item(item_id,id_source,custom_fields,sources=None,verbose=True):
    '''create item will take input for an item, and return the item if it's valid
    :param item_id: the item id to store, mandatory
    :param id_source: mandatory issuer for the above id (e.g., "pacs")
    :param custom_fields: not required, a dictionary of custom key pairs to include as data
    :param id_timestamp: when the id was looked up, to help with changed/merged ids (optional)
    :param sources: a custom list of sources (eg, ['stanford']) to accept. If not defined, defaults
    '''
    item = {"id":item_id,
            "id_source":id_source}

    # Custom fields and id_timestamp are optional
    if custom_fields != None:
        item['custom_fields'] = custom_fields 

    if id_timestamp != None:
        item["id_timestamp"] = id_timestamp

    valid = validate_item(item,sources=sources,
                          custom_fields=custom_fields,
                          verbose=verbose)
    if valid == True:
        return item

    return None



'''
api.google.dlp: the base connection to the google DLP API

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

api_base = "https://dlp.googleapis.com"
api_version = "v2beta1"

from som.logger import bot
from som.api import ApiConnection
from som.api.google.dlp.utils import *

import datetime

from oauth2client.client import GoogleCredentials
from apiclient.discovery import build

class DLPApiConnection(ApiConnection):

    def __init__(self,get_service=True,**kwargs):
        self.token = None
        if 'token' in kwargs:
            self.token = token
        super(DLPApiConnection, self).__init__(**kwargs)
        if get_service is True:
            self.get_service()
        
    def get_service(self):
        credentials = GoogleCredentials.get_application_default()
        self.service = build('dlp', api_version, credentials=credentials)

    def inspect(self,texts,include_quote=True,min_level=None):
        '''inspect will inspect a dump of text for identifiers
        :param include_quote: include quotes in the query?
        :param min_level: the minimum likilihood level to return
        '''
        if not isinstance(texts,list):
            texts = [texts]

        if min_level is None:
            min_level = 'LIKELIHOOD_UNSPECIFIED'

        config = {'includeQuote': include_quote,
                  'infoTypes': [],
                  'maxFindings': 0,
                  'minLikelihood': min_level }
        
        items = []
        for text in texts:
            new_item = {'type': 'text/plain',
                        'value': text }
            items.append(new_item)

        groups = paginate_items(items,size=100)
        
        results = []
        for idx in range(len(groups)):
            bot.debug("inspecting for %s of %s" %(idx+1,len(groups)))
            items = groups[idx]
            body = {'inspectConfig': config,
                    'items': items }
            result = self.service.content().inspect(body=body).execute()
            results = results + result['results']
        return results


    def remove_phi(self,texts,include_quote=True,min_level=None):
        '''remove_phi will first use inspect to find PHI, and then
        prepare the equivalent text with the phi removed.
        :param include_quote: include quotes in the query?
        :param min_level: the minimum likilihood level to return
        '''
        if not isinstance(texts,list):
            texts = [texts]

        results =  self.inspect(texts=texts,
                                include_quote=include_quote,
                                min_level=min_level)

        # We will return an equivalent list with phi removed
        cleaned = []

        for idx in range(len(results)):
            result = results[idx]
            original = texts[idx]
 
            if len(result) is not 0:
                scrubbed = clean_text(text=original,
                                      findings=result)
                cleaned.append(scrubbed)
            else:
                cleaned.append(original)
        return cleaned



'''
google/modals.py: models for datastore

'''

from som.storage.google.validators import (
    validate_model
)

import google.cloud.datastore as datastore
from som.storage.google.datastore import (
    get_key_filters,
    parse_keys,
    print_datastore_path
)

from som.logman import bot
import datetime
import collections
import sys
import os

######################################################################################
# Object Manager
######################################################################################

class BatchManager:
    '''a batch manager is bucket to hold multiple objects to filter, query, etc.
    It a way to compile a set, and then run through a transaction.
    '''

    def __init__(self):
        self.objects = []


    def get_kinds(self,client):
        '''get_kinds'''
        query = client.query(kind='__kind__')
        query.keys_only()
        return [entity.key.id_or_name for entity in query.fetch()]


    def add(self,task):
        '''return all objects in the set
        '''
        if not isinstance(task,datastore.Entity):
            task = task.this
        self.objects.append(task)

        

    def get(self,client,keys):
        '''get will return a list of tasks based on keys, where each
        key is a single key or a model key type (eg "Collection") and
        corresponding identifier Eg:

          keys = ["Collection",["Entity",12345]]

        '''
        with parse_keys(keys) as keys:
            return client.get_multi(keys)



    def delete(self,client,keys):
        '''delete a set of model objects based on user provided keys'''
        with parse_keys(keys) as keys:
            return client.delete_multi(keys) # note sure if returning this 
                                             # returns a status or nothing
                                             # same for get, need to test
                                             # (currently on airplane)


    def insert(self,client,clear_queue=True):
       '''insert will run a transaction for a set of tasks
       '''
       if len(self.objects) > 0:
           tasks = self.objects
           with client.transaction():
               client.put_multi(tasks)
               if clear_queue:
                   self.objects = []
               return tasks
       return None


    def query(self,client,kind=None,filters=None,order=None,projections=None,
              run=True,limit=None,keys_only=False,distinct_on=None,query=None,
              ancestor=None):
        '''query will run a query for some kind of model (eg 'Collection')
        :param kind: the kind of model to query
        :param filters: a list of lists of tuples, each length 3, with 
        :param order: optional one or more orderings -descending / ascending
        :param projections: if defined, return a dictionary of lists of each
        :param run: Default true to execute query. False returns query object
        :param limit: an integer limit of results to return
        :param distinct_on: one or more fields to make distinct
        :param query: if provided, will not start with basic query
        :param ancestor: if provided, query for the provided ancestor
                       
             FIELD OPERATOR VALUE
             "description" "=' "this is a description" 
         
             ANCESTOR EXAMPLE
             ancestor = client.key('TaskList', 'default')

        '''
        if ancestor is not None:
            if isinstance(ancestor,list) or isinstance(ancestor,str):
                ancestor = client.key(*ancestor)

        if query == None:
            if kind is None:
                bot.logger.error("You must define 'kind' to run a query.")
                sys.exit(1)
            query = client.query(kind=kind.upper(),ancestor=ancestor) # uppercase first letter

        if distinct_on != None:
            if not isinstance(distinct_on,list):
                distinct_on = [distinct_on]
            query.distinct_on = ['category', 'priority']

        # Add filters to the query
        if filters is not None:
            for f in filters:
                query.add_filter(f[0], f[1],f[2])

        # Apply order to the query (eg, ['-priority']
        if order is not None:
            if not isinstance(order,list):
                order = [order]
            query.order = order

        # Return a projection, an extraction of fields from a set
        if projections is not None:
            return self.projection(client,query,projections)

        if keys_only:
            query.keys_only()
  
        # Do we not want to run the query?
        if not run:
            return query

        result = None

        try:
            if keys_only:
                result = list([e.key for e in query.fetch(limit=limit)])
            else:
                result = list(query.fetch(limit=limit))

        except (google.cloud.exceptions.BadRequest,
                google.cloud.exceptions.GrpcRendezvous):
            bot.logger.error("Error with query.")         
            pass

        return result


    def query_key(self,client,kind,keys,**kwargs):
        '''query_key is an entry to query that adds a key filter to the query
        '''
        if operator == None:
            operator = '>'

        if operator not in get_key_filters():
            bot.logger.error("%s is not a valid operator.", operator)
            sys.exit(1)

        keys = parse_keys(keys)
        query = client.query(kind=kind)
        for key in keys:
            query.key_filter(key, operator)

        return self.query(client,**kwargs)



    def query_date(self,client,kind,startyear=None,startday=None,startmonth=None,
                   endyear=None,endday=None,endmonth=None,**kwargs):
        '''query_date is an entry to query that sets up a date filter.
        '''
        start_date = None
        end_date=None
        if startyear is not None and startday is not None and startmonth is not None:
            start_date = datetime.datetime(startyear, startmonth, startday)
        if endyear is not None and endmonth is not None and endday is not None:
            end_date = datetime.datetime(endyear, endmonth, endday)
        if start_date == None and end_date == None:
            bot.logger.warning('''Both start date and end date are null, did you provide
                                  year, month, day for one or both?''')
     
        query = client.query(kind=kind)
        if start_date is not None:
            query.add_filter('created', '>', start_date)
        if end_date is not None:
            query.add_filter('created', '<', end_date)
     
        return self.query(client,**kwargs)
        

    def projection(self,client,query,projections,add_loners=False):
        '''extract a subset of a query, meaning a dictionary with key value
        pairs for field/results lists. By default, to preserve indexing,
        only objects with all projection fields are added to the result.
        You can change this behavior by setting add_loners = True
        ''' 
        if not isinstance(projections,list):
            projections = [projections]

        # Find a more elegant way to do this
        results = dict()
        for projection in projections:
            results[projection] = []
                    
        for task in query.fetch():
            loner = False
            for p in projections: 
                if p not in task:
                    longer=True
                    break

            # Find more elegant way to not need to loop twice    
            for p in projections:
                if not loner or (loner and add_loners):
                    results[p].append(task[p])

        return results




######################################################################################
# Base Client Class
######################################################################################

# https://cloud.google.com/datastore/docs/concepts/entities#datastore-basic-entity-python

class ModelBase:

    def __init__(self,client):
        self.fields = validate_model(self.model['fields'])
        self.key = client.key(*self.model['key'])
        if "exclude_from_indexes" in self.model:
            self.exclude_from_indexes = self.model['exclude_from_indexes'] # Can be None
        else:
            self.exclude_from_indexes = None

    def __str__(self):
        return print_datastore_path(self.key)

    def __repr__(self):
        return print_datastore_path(self.key)


    def get_keypath(self):
        return list(self.key.flat_path)
      

    def get_name(self):
        return list(self.key.flat_path)[-1]


    def update_key(self,client,key):
        '''update key will add a key to a model, and return a 
        unique set of keys
        '''
        if key is not None:
           flat_key = list(self.key.flat_path)
           if isinstance(key,list):
               flat_key = flat_key + key
           else: 
               flat_key.append(key)
           self.key = client.key(*flat_key)


    def update_fields(self,new_fields,add_new=True):
        '''update fields will update the model's fields with an input dictionary
        new fields are added if add_new=True (default)
        '''
        for key,value in new_fields.items():
            if key in self.fields:
                self.fields[key] = value
            else:
                if add_new == True:
                    self.fields[key] = value


    def get_or_create(self,client,key=None):
        '''get_or_create will get or create an object using a transaction.
        '''
        with client.transaction():
           self.update_key(client,key)
           entity = client.get(*self.key)

           # The entity is being created, add timestamp
           if not entity:
               entity = self.create(client)
        return entity


    def setup(self,client,fields=None):
        '''setup will generate a new Entity without saving it (for batches)
        '''
        if self.exclude_from_indexes is not None:
            entity = datastore.Entity(key=self.key,
                                      exclude_from_indexes=self.exclude_from_indexes)
        else:
            entity = datastore.Entity(self.key)
        if fields is not None:
            self.fields.update(fields)
        return entity


    def create(self,client,fields=None):
        '''create a new entity
        '''
        entity = self.setup(client,fields)
        self.fields['created'] = datetime.datetime.utcnow()
        self.fields['updated'] = datetime.datetime.utcnow()
        for field,value in self.fields.items():
            entity[field] = value
        client.put(entity)
        return entity


    def delete(self,client):
        '''delete an entity'''
        client.delete(self.key)
        return key


    def update_or_create(self,client,fields=None):
        '''update or create will update or create an entity.
        '''
        entity = client.get(self.key)

        # The entity is being created, add timestamp
        if not entity:
           entity = self.create(client)            #insert
        else: 
            entity = self.update(client,fields) #upsert
        return entity


    def update(self,client,fields=None,key=None):
        '''update an entity. Returns None if does not exist.'''
        entity = None
        with client.transaction():
           entity = client.get(self.key)
           if entity:
               if fields is not None:
                   self.update_fields(fields)

               self.fields['updated'] = datetime.datetime.utcnow()

               for field,value in self.fields.items():
                   entity[field] = value
        return entity


    def get(self,client):
        '''get an entity. Returns None if does not exist.'''
        entity = None
        with client.transaction():
           entity = client.get(self.key)
        return entity



#!/usr/bin/env python

'''
validators/requests.py: validation of json (dict) structures to send

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from som.logger import bot
from som.api.identifiers.standards import (
    identifier_sources,
    item_sources,
    timestamp
)

from .utils import validate_fields

import os
import sys

from validator import (
    Required,
    Pattern, 
    Not, 
    Truthy, 
    Blank, 
    Range, 
    Equals, 
    In, 
    validate
)


def validate_identifiers(identifiers,id_sources=None,item_sources=None,verbose=True):
    '''validate_identifiers will validate one or more identifier objects, 
    intended to go in as a field to a POST
    :param identifiers: the identifiers object.
    :param verbose: verbose output for items
    :param identifier_sources: a list of one or more identifier sources. 
    :param item_sources: a list of one or more item sources
    If either not defined, default to use standards

    :: notes
       {
        # mandatory key for uniquely identifying the person
        "id":"1234567-8",
       
        # the issuer for the above id
        # mandatory, or maybe optional with default of "stanford" or "unspecified"
        
        "id_source":"stanford",
        # when the id was looked up, to help with changed/merged ids
        # optional with default of current timestamp?
        
        "id_timestamp":"2016-01-30T17:15:23.123Z",
        # optional key/value attributes, will be stored as-is if provided, but not used or interpreted
        # values will be updated/replaced if sent multiple times (last received wins)
        # any existing values will be preserved if not explicitly set here; set empty string to remove
        
        "custom_fields":{
          "first_name":"Joe",
          "last_name":"Smith",
          "dob":"1970-02-28"
        }
    '''
    if id_sources == None:
        id_sources = identifier_sources

    # These are the rules for a person
    rules = {
        "id": [Required, Pattern("^[A-Za-z0-9_-]*$")], # pattern
        "id_source": [Required, In(id_sources)],   # must be in person sources
        "id_timestamp": [Required,Pattern(timestamp)], 
    }

    if not isinstance(identifiers,dict):
        bot.error("Identifiers data structure must be dictionary.")
        return False

    if "identifiers" not in identifiers:
        bot.error("identifiers key not found in data structure.")

    items = identifiers['identifiers']

    if not isinstance(items,list):
        bot.error("Items in identifiers data structure must be list.")
        return False  

    for item in items:

        valid,message = validate(rules, item)
        if valid == False:
            bot.error(message)
            return valid

        if "items" in item:
            validate_items(item['items'],sources=item_sources)

    bot.debug("Identifiers data structure valid: %s" %valid)
    return valid


def validate_items(items,sources=None,verbose=True):
    '''validate_items is a wrapper for "validate_item" to allow for one or more items
    to go through validation at once.
    :param items: one or more items to parse through, a dict or list of dicts
    :param item_sources: custom list of item sources. If not provided, defaults used.
    :param verbose: if True, will output each item that is valid
    '''
    if not isinstance(items,list):
        items = [items]

    for item in items:
        valid = validate_item(item=item,
                              sources=sources,
                              verbose=verbose)
 
        # We stop mid point if we find an invalid item
        if valid == False:
            return valid
 
    # Otherwise, we return all items valid
    return valid


def validate_item(item,sources=None,verbose=True):
    '''validate_item will validate a single item objects, intended to go in as a field to
    a POST. For more than one item, use validate_items wrapper
    :param item: the item object. Must include the following:
    :param sources: a list of valid item sources (eg ["pacs"])
    :param verbose: if True, prints out valid True messages. False (errors) always printed

    :: notes

    {
        # generic attribute/values just to store, not interpreted
        "id":"123123123123123", // mandatory

        # the issuer for the above id
        # mandatory, or maybe optional with default of "stanford" or "unspecified"
        "id_source":"pacs",

        # when the id was looked up, to help with changed/merged ids
        # optional with default of current timestamp?
        "id_timestamp":"2016-01-30T17:15:23.123Z",

        # optional key/value attributes, will be stored as-is if provided, but not used or interpreted
        # values will be updated/replaced if sent multiple times (last received wins)
        # any existing values will be preserved if not explicitly set here; set empty string to remove
        "custom_fields":{
          "image_type":"x-ray",
          "resolution":"high"
    }
    '''
    if sources == None:
        sources = item_sources

    # These are the rules for an item
    rules = {
        "id": [Required, Pattern("^[A-Za-z0-9_-]*$")], # pattern
        "id_source": [Required, In(sources)],      # must be in item sources
        "id_timestamp": [Required,Pattern(timestamp)], 
    }

    valid,message = validate(rules, item)
    if verbose == True:
        bot.debug("identifier %s data structure valid: %s" %(item['id'],valid))
    if valid == False:
        bot.error(message)
        if verbose == True:
            print(item)
    return valid



#!/usr/bin/env python

'''
base.py: base module for working with som api

'''

from som.storage.google.client import ClientBase
from som.storage.google.models import *
from som.storage.google.utils import *
from som.logman import bot
from som.utils import read_json
from som.wordfish.structures import structure_dataset
import six

######################################################################################
# Specs and base data structures for each radiology model
######################################################################################


def entity(uid,collection,metadata=None):
    '''entity returns an entity object
    parent is a collection
    '''
    fields = [{'key':'uid','required':True,'value':uid},
              {'key':'metadata','required':False,'value':metadata}]

    if type(collection) not in six.string_types:
        collection = collection.get_name()
    model = {'fields':fields,
             'key':['Collection',collection,'Entity', uid]}

    return model



def collection(uid,metadata=None,**fields):
    '''entity returns an entity object
    parent is an owner
    '''
    fields = [{'key':'uid','required':True,'value':uid},
              {'key':'metadata','required':False,'value':metadata}]

    model = {'fields':fields,
             'key':['Collection', uid]}
    return model
    


def image(uid,entity,download,url,storage,metadata=None,**fields):
    '''image returns an image object. entity is the parent
    '''
    fields = [{'key':'uid','required':True,'value':uid},
              {'key':'metadata','required':False,'value':metadata},
              {'key':'download','required':True,'value':download,
               'key':'url','required':True,'value': url,
               'key':'storage','required':True,'value': storage}]

    collection = entity.collection.get_name()
    entity = entity.get_name()

    model = {'fields':fields,
             'exclude_from_indexes': ['storage','download','url'],
             'key':['Collection', collection, 'Entity',entity,'Image', uid]}

    return model
    


def text(uid,entity,content,metadata=None,**fields):
    '''text returns a text object. entity is the parent
    '''
    fields = [{'key':'uid','required':True,'value':uid},
              {'key':'metadata','required':False,'value':metadata},
              {'key':'content','required':True,'value':content}]

    collection = entity.collection.get_name()
    entity = entity.get_name()

    model =  {'fields':fields,
              'exclude_from_indexes': ['storage','download','url'],
              'key':['Collection', collection, 'Entity',entity,'Text', uid]}
    return model




######################################################################################
# Core Models, extending base model
######################################################################################


class Collection(ModelBase):
  
    def __init__(self,client,uid,create=True,**fields):
        self.model = collection(uid=uid,**fields)
        super(Collection, self).__init__(client,**fields)
        if create:
            self.this = self.update_or_create(client)
        else:
            self.this = self.setup(client)


class Entity(ModelBase):

    def __init__(self,client,collection,uid,create=True,**fields):
        self.collection = collection
        self.model = entity(uid=uid, collection=collection,**fields)
        super(Entity, self).__init__(client,**fields)
        if create:
            self.this = self.update_or_create(client)
        else:
            self.this = self.setup(client)
        

    def collection(self,client):
        '''collection will return the collection associated with
        the entity
        '''
        return client.get(collection.key)


    def images(self,client):
        '''images will return images associated with the entity
        '''
        key = self.get_keypath()
        return client.get(client.key(*key, "Image"))
    

    def text(self,client):
        '''text will return text associated with the entity
        '''
        key = self.get_keypath()
        return client.get(client.key(*key, "Text"))


class Image(ModelBase):

    def __init__(self,client,uid,entity,create=True,url=None,
                 download=None,storage=None,**fields):
        self.entity = entity
        self.model = image(uid=uid,entity=entity,url=url,
                           storage=storage,download=download,**fields)
        super(Image, self).__init__(client,**fields)
        if create:
            self.this = self.update_or_create(client)
        else:
            self.this = self.setup(client)

    def __repr__(self):
        return self.this


class Text(ModelBase):

    def __init__(self,client,uid,entity,content,create=True,**fields):
        self.entity = entity
        self.model = text(uid=uid,entity=entity,content=content,**fields)

        super(Text, self).__init__(client,**fields)
        if create:
            self.this = self.get_or_create(client)
        else:
            self.this = self.setup(client)


    def __repr__(self):
        return self.this




######################################################################################
# Radiology Client to interact with Models
######################################################################################


class Client(ClientBase):

    def __init__(self,**kwargs):
        self.bucket_name = 'radiology'
        self.batch = BatchManager()
        super(Client, self).__init__(**kwargs)
    
    def __str__(self):
        return "storage.google.%s" %self.bucket_name

    def __repr__(self):
        return "storage.google.%s" %self.bucket_name


    ###################################################################
    ## Create and get #################################################
    ###################################################################

    def get_collection(self,fields):
        return Collection(client=self.datastore,**fields)

    def get_entity(self,fields):
        return Entity(client=self.datastore,**fields)


    def create_image(self,fields):
        return Image(client=self.datastore,**fields)

    def create_text(self,fields):
        return Text(client=self.datastore,**fields)

    def get_storage_path(self,file_path,collection_name,return_folder=False):
        folder = file_path.split(collection_name)[1]
        bucket_path = "%s%s" %(collection_name,
                               folder)
        if return_folder:
            return os.path.dirname(bucket_path)
        return bucket_path



    ###################################################################
    ## Upload #########################################################
    ###################################################################

    def upload_texts(self,texts,entity,batch=True):
        '''upload_texts will add text objects to the batch manager'''

        if not isinstance(texts,list):
            texts = [texts]

        new_texts = []
        for txt in texts:
            uid = self.get_storage_path(txt['original'],entity.collection.get_name())

            with open(txt['original'],'r') as filey:
                content = filey.read()

            fields = {'uid':uid,
                      'entity':entity,
                      'content':content,
                      'create': not batch }

            if 'metadata' in txt:
                fields['metadata'] = read_json(txt['metadata'])
 
            new_text = self.create_text(fields) # Create if not doing batch
            if batch:
                self.batch.add(new_text)

            bot.logger.debug('TEXT: %s',new_text)
            new_texts.append(new_text)
        return new_texts


    def upload_images(self,images,entity,batch=True):
        '''upload_images will add image objects to the batch manager, or create
        one at a time
        '''
        if not isinstance(images,list):
            images = [images]

        new_images = []
        collection_name = entity.collection.get_name()

        for img in images:
            bucket_folder = self.get_storage_path(img['original'],collection_name)
            image_obj = self.upload_object(file_path=img['original'],
                                           bucket_folder=bucket_folder)

            url = self.get_storage_path(image_obj['name'],collection_name)

            fields = {'uid':image_obj['id'],
                      'entity':entity,
                      'url':url,
                      'storage':image_obj,
                      'download':image_obj['mediaLink'],
                      'create': not batch }

            new_image = self.create_image(fields)
            new_images.append(new_image)

            bot.logger.debug('IMAGE: %s',new_image)

            if batch:
                self.batch.add(new_image)

        return new_images


    def upload_dataset(self,structures,batch=True):
        '''upload takes a list of structures (eg from som.wordfish.structures
        and uploads to storage and datastore. 
        :param structures: the list of wordfish data structures
        :param batch: add entities in batches (recommended, default True)
        '''
        
        if not isinstance(structures,list):
            structures = [structures]

        # Each structure corresponds to a collection, a subset of entities
        for coll in structures:

            # The structure should be created first
            uid = os.path.basename(coll['collection']['name'])
            fields = {'uid': uid }
            if 'metadata' in coll['collection']:
                fields['metadata'] = read_json(coll['collection']['metadata'])        
            new_collection = self.get_collection(fields)
            bot.logger.debug('GET: %s',new_collection)

            # Add entities
            if 'entities' in coll['collection']:
                for ntt in coll['collection']['entities']:
                    
                    fields = {'uid': os.path.basename(ntt['entity']['id']),
                              'collection':new_collection }
                    entity = self.get_entity(fields)
                   
                    # Add images and text to entities
                    if 'texts' in ntt['entity']:
                        new_texts = self.upload_texts(texts=ntt['entity']['texts'],
                                                      entity=entity,
                                                      batch=batch)

                    if 'images' in ntt['entity']:
                        new_images = self.upload_images(images=ntt['entity']['images'],
                                                        entity=entity,
                                                        batch=batch)

                    # Run a transaction for put (insert) images and text, and clears queue
                    self.batch.insert(client=self.datastore)




def cursor_paging(client):
    # [START cursor_paging]

    def get_one_page_of_tasks(cursor=None):
        query = client.query(kind='Task')
        query_iter = query.fetch(start_cursor=cursor, limit=5)
        page = next(query_iter.pages)

        tasks = list(page)
        next_cursor = query_iter.next_page_token

        return tasks, next_cursor
    # [END cursor_paging]

    page_one, cursor_one = get_one_page_of_tasks()
    page_two, cursor_two = get_one_page_of_tasks(cursor=cursor_one)
    return page_one, cursor_one, page_two, cursor_two





def transactional_single_entity_group_read_only(client):
    client.put_multi([
        datastore.Entity(key=client.key('TaskList', 'default')),
        datastore.Entity(key=client.key('TaskList', 'default', 'Task', 1))
    ])

    # [START transactional_single_entity_group_read_only]
    with client.transaction():
        task_list_key = client.key('TaskList', 'default')

        task_list = client.get(task_list_key)

        query = client.query(kind='Task', ancestor=task_list_key)
        tasks_in_list = list(query.fetch())

        return task_list, tasks_in_list
    # [END transactional_single_entity_group_read_only]


def namespace_run_query(client):
    # Create an entity in another namespace.
    task = datastore.Entity(
        client.key('Task', 'sample-task', namespace='google'))
    client.put(task)

    # [START namespace_run_query]
    # All namespaces
    query = client.query(kind='__namespace__')
    query.keys_only()

    all_namespaces = [entity.key.id_or_name for entity in query.fetch()]

    # Filtered namespaces
    start_namespace = client.key('__namespace__', 'g')
    end_namespace = client.key('__namespace__', 'h')
    query = client.query(kind='__namespace__')
    query.key_filter(start_namespace, '>=')
    query.key_filter(end_namespace, '<')

    filtered_namespaces = [entity.key.id_or_name for entity in query.fetch()]
    # [END namespace_run_query]

    return all_namespaces, filtered_namespaces




def property_run_query(client):
    # Create the entity that we're going to query.
    upsert(client)

    # [START property_run_query]
    query = client.query(kind='__property__')
    query.keys_only()

    properties_by_kind = defaultdict(list)

    for entity in query.fetch():
        kind = entity.key.parent.name
        property_ = entity.key.name

        properties_by_kind[kind].append(property_)
    # [END property_run_query]

    return properties_by_kind


def property_by_kind_run_query(client):
    # Create the entity that we're going to query.
    upsert(client)

    # [START property_by_kind_run_query]
    ancestor = client.key('__kind__', 'Task')
    query = client.query(kind='__property__', ancestor=ancestor)

    representations_by_property = {}

    for entity in query.fetch():
        property_name = entity.key.name
        property_types = entity['property_representation']

        representations_by_property[property_name] = property_types
    # [END property_by_kind_run_query]

    return representations_by_property


def eventual_consistent_query(client):
    # [START eventual_consistent_query]
    # Read consistency cannot be specified in google-cloud-python.
    # [END eventual_consistent_query]
    pass


def main(project_id):
    client = datastore.Client(project_id)

    for name, function in globals().iteritems():
        if name in ('main', 'defaultdict') or not callable(function):
            continue

        print(name)
        pprint(function(client))
        print('\n-----------------\n')






# coding: utf-8

'''

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from __future__ import absolute_import

import os
import sys
import unittest

from som.api.radiology import Client
from som.api.validators.responses import receive_identifiers
from som.api.validators.requests import validate_identifiers
from som.utils import read_json

here = os.path.dirname(os.path.realpath(__file__))

class TestDevelopersApi(unittest.TestCase):
    """ DevelopersApi unit tests"""


    def setUp(self):
        self.client = Client()
        self.identifiers = read_json('%s/data/developers_uid.json' %(here))


    def tearDown(self):
        pass

    def test_uid(self):
        '''Test case for uid
        Accepts a list of identified items, returns a list of study specific identifiers
        '''

        print("Case 1: Formatting of identifiers input is correct.")
        self.assertTrue(validate_identifiers(self.identifiers))

        print("Case 2: Respones returns status 200, json object")
        response = self.client.deidentify(self.identifiers)
        self.assertTrue(isinstance(response,list))       

        # Validate response
        print("Case 3: Formatting of identifiers response is correct.")
        self.assertTrue(receive_identifiers(response))

        # Assert we have the same number of inputs and outputs
        print("Case 4: Response returns same number of identifiers as given, %s" %(len(response)))
        self.assertEqual(len(response),len(self.identifiers['identifiers']))


if __name__ == '__main__':
    unittest.main()



'''
google/utils.py: general storage utility functions

'''


from googleapiclient.errors import HttpError
from googleapiclient import (
    http,
    discovery
)

from oauth2client.client import GoogleCredentials
from oauth2client.service_account import ServiceAccountCredentials

from som.logman import bot

from som.utils import (
    get_installdir,
    read_file,
    run_command
)

from subprocess import (
    Popen,
    PIPE,
    STDOUT
)

import tempfile
import zipfile

from glob import glob
import httplib2
import inspect
import imp
import os
import json
import pickle
import re
import requests
from retrying import retry
import shutil
import simplejson
import sys
import tempfile
import time
import zipfile


######################################################################################
# Testing/Retry Functions
######################################################################################

def stop_if_result_none(result):
    '''stop if result none will return True if we should not retry 
    when result is none, False otherwise using retrying python package
    '''
    do_retry = result is not None
    return do_retry



##########################################################################################
# GOOGLE GENERAL API #####################################################################
##########################################################################################

def get_google_service(service_type=None,version=None):
    '''
    get_url will use the requests library to get a url
    :param service_type: the service to get (default is storage)
    :param version: version to use (default is v1)
    '''
    if service_type == None:
        service_type = "storage"
    if version == None:
        version = "v1"

    credentials = GoogleCredentials.get_application_default()
    return discovery.build(service_type, version, credentials=credentials) 


##########################################################################################
# GOOGLE STORAGE API #####################################################################
##########################################################################################
    
def get_bucket(storage_service,bucket_name):
    req = storage_service.buckets().get(bucket=bucket_name)
    return req.execute()


def delete_object(storage_service,bucket_name,object_name):
    '''delete_file will delete a file from a bucket
    :param storage_service: the service obtained with get_storage_service
    :param bucket_name: the name of the bucket (eg singularity-hub)
    :param object_name: the "name" parameter of the object.
    '''
    try:
        operation = storage_service.objects().delete(bucket=bucket_name,
                                                     object=object_name).execute()
    except HttpError as e:
        pass
        operation = e
    return operation


def upload_file(storage_service,bucket,bucket_folder,file_path,verbose=True):
    '''get_folder will return the folder with folder_name, and if create=True,
    will create it if not found. If folder is found or created, the metadata is
    returned, otherwise None is returned
    :param storage_service: the drive_service created from get_storage_service
    :param bucket: the bucket object from get_bucket
    :param file_path: the name of the file to upload (local)
    :param bucket_folder: the "folder" path to upload to
    '''
    # Set up path on bucket
    upload_path = bucket_folder
    if upload_path[-1] != '/':
        upload_path = "%s/" %(upload_path)
    upload_path = "%s%s" %(upload_path,os.path.basename(file_path))
    body = {'name': upload_path }

    # Create media object with correct mimetype
    mimetype = sniff_extension(file_path,verbose=verbose)
    media = http.MediaFileUpload(file_path,
                                 mimetype=mimetype,
                                 resumable=True)
    request = storage_service.objects().insert(bucket=bucket['id'], 
                                               body=body,
                                               predefinedAcl="publicRead",
                                               media_body=media)
    return request.execute()



def list_bucket(bucket,storage_service):
    # Create a request to objects.list to retrieve a list of objects.        
    request = storage_service.objects().list(bucket=bucket['id'], 
                                             fields='nextPageToken,items(name,size,contentType)')
    # Go through the request and look for the folder
    objects = []
    while request:
        response = request.execute()
        objects = objects + response['items']
    return objects


def get_image_path(repo_url,commit):
    '''get_image_path will determine an image path based on a repo url, removing
    any token, and taking into account urls that end with .git.
    :param repo_url: the repo url to parse:
    :param commit: the commit to use
    '''
    repo_url = repo_url.split('@')[-1].strip()
    if repo_url.endswith('.git'):
        repo_url =  repo_url[:-4]
    return "%s/%s" %(re.sub('^http.+//www[.]','',repo_url),commit)




#####################################################################################
# METADATA
#####################################################################################


def get_metadata(key):
    '''get_metadata will return metadata about an instance from within it.
    :param key: the key to look up
    '''
    headers = {"Metadata-Flavor":"Google"}
    url = "http://metadata.google.internal/computeMetadata/v1/instance/attributes/%s" %(key)        
    response = api_get(url=url,headers=headers)

    # Successful query returns the result
    if response.status_code == 200:
        return response.text
    else:
        bot.logger.error("Error retrieving metadata %s, returned response %s", key,
                                                                               response.status_code)
    return None



######################################################################################
# Extensions and Files
######################################################################################


def sniff_extension(file_path,verbose=True):
    '''sniff_extension will attempt to determine the file type based on the extension,
    and return the proper mimetype
    :param file_path: the full path to the file to sniff
    :param verbose: print stuff out
    '''
    mime_types =    { "xls": 'application/vnd.ms-excel',
                      "xlsx": 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
                      "xml": 'text/xml',
                      "ods": 'application/vnd.oasis.opendocument.spreadsheet',
                      "csv": 'text/plain',
                      "tmpl": 'text/plain',
                      "pdf":  'application/pdf',
                      "php": 'application/x-httpd-php',
                      "jpg": 'image/jpeg',
                      "png": 'image/png',
                      "gif": 'image/gif',
                      "bmp": 'image/bmp',
                      "txt": 'text/plain',
                      "doc": 'application/msword',
                      "js": 'text/js',
                      "swf": 'application/x-shockwave-flash',
                      "mp3": 'audio/mpeg',
                      "zip": 'application/zip',
                      "rar": 'application/rar',
                      "tar": 'application/tar',
                      "arj": 'application/arj',
                      "cab": 'application/cab',
                      "html": 'text/html',
                      "htm": 'text/html',
                      "default": 'application/octet-stream',
                      "folder": 'application/vnd.google-apps.folder',
                      "img" : "application/octet-stream" }

    ext = os.path.basename(file_path).split('.')[-1]

    mime_type = mime_types.get(ext,None)

    if mime_type == None:
        mime_type = mime_types['txt']

    if verbose==True:
        bot.logger.info("%s --> %s", file_path, mime_type)

    return mime_type

    



'''

som.api: base template for making a connection to an API

'''

from som.logger import bot
import requests
import json
import sys
import re
import os


class ApiConnection(object):

    def __init__(self, token=None):
 
        self.headers = None
        if token is not None:
            self.token = token
        self.update_headers()
        

    def get_headers(self):
        return self.headers

    def _init_headers(self):
        return {'Content-Type':"application/json"}


    def update_headers(self,fields=None):
        '''get_headers will return a simple default header for a json
        post. This function will be adopted as needed.
        '''
        if self.headers is None:
            headers = self._init_headers()
        else:
            headers = self.headers

        if self.token is not None:
            headers["Authorization"] = "Bearer %s" %(self.token)

        if fields is not None:
            for key,value in fields.items():
                headers[key] = value

        header_names = ",".join(list(headers.keys()))
        bot.debug("Headers found: %s" %header_names)
        self.headers = headers


    def put(self,url,data=None,return_json=True):
        '''put request
        '''
        bot.debug("PUT %s" %url)
        return self.call(url,
                         func=requests.put,
                         data=data,
                         return_json=return_json)



    def post(self,url,data=None,return_json=True):
        '''post will use requests to get a particular url
        '''
        bot.debug("POST %s" %url)
        return self.call(url,
                        func=requests.post,
                        data=data,
                        return_json=return_json)



    def get(self,url,headers=None,token=None,data=None, return_json=True):
        '''get will use requests to get a particular url
        '''
        bot.debug("GET %s" %url)
        return self.call(url,
                        func=requests.get,
                        data=data,
                        return_json=return_json)



    def call(self,url,func,data=None,return_json=True):
        '''call is a template that will take a post, put, or get function
        and insert the right variables.
        '''
        response = func(url,headers=self.headers,
                            data=json.dumps(data))

        if response.status_code == 200 and return_json:
            return response.json()

        return response



#!/usr/bin/env python

'''
general.py: general models for text and images

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from som.api.google.storage.client import ClientBase
from som.api.google.storage.models import *
from som.api.google.storage.utils import *
from som.logger import bot
from som.utils import read_json
import six

######################################################################################
# Specs and base data structures for a general model
######################################################################################


def entity(uid,collection):
    '''entity returns an entity object
    parent is a collection
    '''
    fields = [{'key':'uid','required':True,'value':uid}]

    if type(collection) not in six.string_types:
        collection = collection.get_name()
    model = {'fields':fields,
             'key':['Collection',collection,'Entity', uid]}

    return model



def collection(uid):
    '''entity returns an entity object
    parent is an owner
    '''
    fields = [{'key':'uid','required':True,'value':uid}]

    model = {'fields':fields,
             'key':['Collection', uid]}
    return model
    


def storageObject(uid,entity,url,storage_type):
    '''image returns an image object. entity is the parent
    '''
    fields = [{'key':'uid','required':True,'value': uid},
              {'key':'url','required':True,'value': url}]

    collection = entity.collection.get_name()
    entity = entity.get_name()

    storage_type = storage_type.replace(' ','-').lower().capitalize()

    model = {'fields':fields,
             'exclude_from_indexes': ['url'],
             'key':['Collection', collection, 'Entity',entity, storage_type, uid]}

    return model
    


######################################################################################
# Core Models, extending base model
######################################################################################


class Collection(ModelBase):
  
    def __init__(self,client,uid,create=True,fields=None):
        self.model = collection(uid=uid)
        super(Collection, self).__init__(client)
        if create:
            self.update_or_create(fields=fields)
        else:
            self.update_or_create(fields=fields,
                                  save=False)


class Entity(ModelBase):

    def __init__(self,client,collection,uid,create=False,fields=None):
        self.collection = collection
        self.model = entity(uid,collection)
        super(Entity, self).__init__(client)
        if create:
            self.update_or_create(fields=fields)
        else:
            self.update_or_create(fields=fields,
                                  save=False)
        

class Object(ModelBase):

    def __init__(self,client,uid,entity,url,object_type,create=True,fields=None):
        self.entity = entity
        self.model = storageObject(uid=uid,entity=entity,url=url,storage_type=object_type)
        super(Object, self).__init__(client)
        if create:
            self.update_or_create(fields=fields)
        else:
            self.update_or_create(fields=fields,
                                  save=False)



######################################################################################
# Client to interact with Models
######################################################################################


class Client(ClientBase):

    def __init__(self,bucket_name,**kwargs):
        self.bucket_name = bucket_name
        super(Client, self).__init__(**kwargs)
    
    def __str__(self):
        return "storage.google.%s" %self.bucket_name

    def __repr__(self):
        return "storage.google.%s" %self.bucket_name



    ###################################################################
    ## Create #########################################################
    ###################################################################

    def create_collection(self,uid,create=True,fields=None):
        return Collection(client=self.datastore,
                          uid=uid,
                          create=create,
                          fields=fields)


    def create_entity(self,collection,uid,create=True,fields=None):
        return Entity(client=self.datastore,
                      collection=collection,
                      uid=uid,
                      create=create,
                      fields=fields)


    def create_object(self,uid,entity,url,object_type,create=True,fields=None):
        '''Object type should be one in Image or Text'''
        return Object(client=self.datastore,
                      object_type=object_type,
                      uid=uid,
                      entity=entity,
                      url=url,
                      create=create,
                      fields=fields)


    def create_object(self,uid,entity,url,object_type,create=True,fields=None):
        '''Object type should be one in Image or Text'''
        return Object(client=self.datastore,
                      object_type=object_type,
                      uid=uid,
                      entity=entity,
                      url=url,
                      create=create,
                      fields=fields)

    ###################################################################
    ## Get ############################################################
    ###################################################################

    def get_storage_path(self,file_path,entity,return_folder=False):
        folder = '/'.join(entity.get_keypath())
        bucket_path = "%s/%s" %(folder,os.path.basename(file_path))
        if return_folder:
            return os.path.dirname(bucket_path)
        return bucket_path


    def get_collections(self,uids=None,limit=None,keys_only=False):
        return self.batch.get(kind="Collection",
                              keys=uids,
                              limit=limit,
                              keys_only=keys_only)


    def get_entities(self,collection=None,field=None,uids=None,limit=None,keys_only=False):
        '''eg:     pmc_articles = client.get_entities(uids=pmc_keys,field="pmcid")
        '''  
        ancestor = None
        if collection is not None:
            ancestor = self.batch.client.key("Collection", collection) 
        return self.batch.get(kind="Entity",
                              limit=limit,
                              field=field,
                              ancestor=ancestor,
                              keys=uids,
                              keys_only=keys_only)


    def get_images(self,entity,limit=None,keys_only=False):
        return self.batch.query(kind="Image",
                                limit=limit,
                                keys_only=keys_only,
                                ancestor=entity.key)


    def get_text(self,entity,limit=None,keys_only=False):
        return self.batch.query(kind="Text",
                                limit=limit,
                                keys_only=keys_only,
                                ancestor=entity.key)

  
    ###################################################################
    ## Upload #########################################################
    ###################################################################

    def upload_object(self,file_path,entity,
                      object_type=None,
                      batch=True,
                      fields=None):

        '''upload_object will add a general object to the batch manager
        The object is uploaded to Google Storage, returning storage fields.
        If the user has provided additional fields, these are added to the
        call to create a new object (datastore)'''

        if object_type is None:
            bot.warning("You must specify object_type. Image or Text.")
            return None

        uid = self.get_storage_path(file_path,entity)
        bucket_folder = self.get_storage_path(file_path,entity,return_folder=True)

        storage_obj = self.put_object(file_path=file_path,
                                      bucket_folder=bucket_folder)

        # Obtain storage fields, update with provided fields
        storage_fields = get_storage_fields(storage_obj)
        if fields is not None:
            storage_fields.update(fields)
        fields = storage_fields

        url = "https://storage.googleapis.com/%s/%s" %(self.bucket['name'],
                                                       storage_obj['name'])

        new_object = self.create_object(uid=uid,
                                        entity=entity,
                                        url=url,
                                        fields=fields,
                                        object_type=object_type,
                                        create=not batch)

        if batch:
            self.batch.add(new_object)
        return new_object


    def upload_text(self,text,entity,batch=True,fields=None):
        '''upload_text will add a text object to the batch manager'''
        new_object = self.upload_object(file_path=text,
                                        entity=entity,
                                        fields=fields,
                                        object_type="Text",
                                        batch=batch)
        bot.debug('TEXT: %s' %new_object)
        return new_object


    def upload_image(self,image,entity,batch=True,fields=None):
        '''upload_images will add an image object to the batch manager
        '''
        new_object = self.upload_object(file_path=image,
                                        entity=entity,
                                        fields=fields,
                                        object_type="Image",
                                        batch=batch)
        bot.debug('IMAGE: %s' %new_object)
        return new_object


    def upload_dataset(self,uid,collection,
                       images=None,
                       texts=None,
                       entity_metadata=None,
                       images_metadata=None,
                       texts_metadata=None,
                       batch=True):

        '''upload takes a list of images, texts, and optional metadata
        and uploads to datastore (metadata) and storage (images)
        :param uid: should be the unique id for the entity, with one or more images/texts
        :param entity_metadata: a single dictionary of keys/values for the entity
        :param texts_metadata: a dictionary with keys corresponding to text paths
        of key value pairs for the text in question
        :param images_metadata: the same, but for images
        :param collection: should be the collection to add the entity to
        :param batch: add entities in batches (recommended, default True)
        '''

        if images_metadata is None: images_metadata = {}
        if texts_metadata is None: texts_metadata = {}

        # Add entity
        entity = self.create_entity(collection=collection,
                                    uid=uid)

        if entity_metadata is not None:
            entity.update(fields=entity_metadata)
       
        if texts is not None:

            for text in texts:
                fields = None

                # metadata provided for the text?
                if text in texts_metadata:
                    fields = texts_metadata[text]

                self.upload_text(text=text,
                                 entity=entity,
                                 fields=fields,
                                 batch=batch)



        if images is not None:
            for img in images:
                fields = None

                # metadata provided for the image?
                if img in images_metadata:
                    fields = images_metadata[img]

                self.upload_image(image=img,
                                  entity=entity,
                                  fields=fields,
                                  batch=batch)

        # Run a transaction for put (insert) images and text, and clears queue
        if batch:
            self.batch.runInsert()



'''
utils.py: helper functions for working with dicom module

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


import dateutil.parser
from som.api.identifiers.standards import valid_actions
from som.logger import bot
from pydicom import read_file
from pydicom._dicom_dict import DicomDictionary
from pydicom.tag import Tag


#########################################################################
# Functions for Single Dicom files
#########################################################################

def add_tag(dicom,name,value):
    '''add tag will add a tag only if it's in the (active) DicomDictionary
    :param dicom: the pydicom.dataset Dataset (pydicom.read_file)
    :param name: the name of the field to add
    :param value: the value to set, if name is a valid tag
    '''
    dicom_file = os.path.basename(dicom.filename)
    tag = get_tag(name)

    if name in tag:
        dicom.add_new(tag['tag'], tag['VR'], value) 
 
        # dicom.data_element("PatientIdentityRemoved")
        # (0012, 0062) Patient Identity Removed            CS: 'Yes'

        bot.debug("ADDITION %s to %s." %(dicom.data_element(name),dicom_file))
    else:
        bot.error("%s is not a valid field to add. Skipping." %(name))

    return dicom


def get_tag(name):
    '''get_tag will return a dictionary with tag indexed by name. For each entry,
    a dictionary lookup is included with VR,  
    :name: the keyword to get tag for, eg "PatientIdentityRemoved"
    '''
    found = [{key:value} for key,value in DicomDictionary.items() if value[4] == name]
    tags = dict()
    if len(found) > 0:

        # (VR, VM, Name, Retired, Keyword
        found = found[0] # shouldn't ever have length > 1
        tag = Tag(list(found)[0])
        VR, VM, longName, retired, keyword = found[tag]

        manifest = {"tag": tag,
                    "VR": VR,
                    "VM":VM,
                    "keyword":keyword,
                    "name":longName }

        tags[name] = manifest 
    return tags


def update_tag(dicom,name,value):
    '''update tag will update a value in the header, if it exists
    if not, nothing is added. If the user wants to add a value
    (that might not exist) the function add_tag should be used
    '''
    if name in dicom:
        tag = dicom.data_element(name).tag
        dicom[tag] = value
    return dicom


def blank_tag(dicom,name):
    '''blank tag calls update_tag with value set to an
    empty string.
    '''
    return update_tag(dicom,name,"")


def remove_tag(dicom,name):
    '''remove tag will remove a tag if it is present in the dataset
    :param dicom: the pydicom.dataset Dataset (pydicom.read_file)
    :param name: the name of the field to remove
    '''
    if name in dicom:
        tag = dicom.data_element(name).tag
        del dicom[tag]
    return dicom


#########################################################################
# Config.json Helpers
#########################################################################


def get_func(function_name):
    '''get_func will return a function that is defined from a string.
    the function is assumed to be in this file
    '''
    env = globals()
    if function_name in env:
        return env[function_name]
    return None


def perform_addition(config,dicom):
    '''perform addition will add any additional fields to the dicom
    :param config: the config, with expected ['response']['additions']
    :param dicom: the dicom file to add a field to
    if the field is already in the dicom, it will be overwritten
    '''
    ## Actions/Additions for each come from config
    additions = config['response']['additions']

    for addition in additions:

        name = addition['name']
        value = addition['value']
        dicom = add_tag(dicom,name,value) 

    return dicom


def perform_action(dicom,response,params):
    '''perform action takes  
    :param dicom: a loaded dicom file (pydicom read_file)
    and a dictionary of params.
    :param response: the api response, with "id", "id_source" "items"
    at the top level
    :param params:
        "name" (eg, PatientID) the header field to process
        "action" (eg, coded) what to do with the field
        "value": if needed, the field from the response to replace with
    '''
    header = entity_action.get('name')   # e.g: PatientID
    action = entity_action.get('action') # "coded"
    value = entity_action.get('value')   # "suid"
    done = False

    if action not in valid_actions:
        bot.warning('%s in not a valid choice [%s]. Defaulting to blanked.' %(action,
                                                                              ".".join(valid_actions)))
        action = "blanked"

    if header in dicom:

        # Blank the value
        if action == "blanked":
            dicom = blank_tag(dicom,name=header)
            done = True
 
        # Code the value with something in the response
        elif entity_action == "coded":
            if value in response:
                dicom = update_tag(dicom,
                                   name=header,
                                   value=response[value])
                done = True

        # Remove the field entirely
        elif entity_action == "removed":
            dicom = remove_tag(dicom,name=header)
            done = True

        # Do nothing. Keep the original
        elif entity_action == "original":
            done = True

        if not done:            
            dicom = blank_tag(dicom,name=header)

    return dicom


# Timestamps

def get_item_timestamp(dicom):
    '''get_item_timestamp will return the UTC time for an instance.
    This is derived from the InstanceCreationDate and InstanceCreationTime
    If the Time is not set, only the date is used.
    # testing function https://gist.github.com/vsoch/23d6b313bd231cad855877dc544c98ed
    '''
    item_time = dicom.get("InstanceCreationTime","")
    item_date = dicom.get("InstanceCreationDate")
    timestamp = dateutil.parser.parse("%s%s" %(item_date,item_time))
    return timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")


def get_entity_timestamp(dicom):
    '''get_entity_timestamp will return a UTC timestamp for the entity,
    derived from the patient's birthdate. In the config.json, this is
    set by setting type=func, and value=get_entity_timestamp
    '''
    item_date = dicom.get("PatientBirthDate")
    timestamp = dateutil.parser.parse("%s" %(item_date))
    return timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")




#!/usr/bin/env python

'''

dicom.py: Helper functions for working with dicoms, not finished developing

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


import dicom
import nibabel
import numpy
from glob import glob
import os
from som.logger import bot


def find_dicoms(folder,extension=None):
    '''find_dicoms will walk a directory structure and find dicoms in subfolders
    :param folder: the parent folder to look in
    :param extension: the extension to use. Default is .dcm
    '''
    folders = dict()

    if extension == None:
        extension = ".dcm"

    for dirpath, dirnames, filenames in os.walk(folder):
        dicoms = []
        for filename in [f for f in filenames if f.endswith(extension)]:
            dicoms.append(os.path.join(dirpath, filename))
        if len(dicoms) > 0:
            folders[dirpath] = dicoms
    bot.debug('Found %s directories with dicom.' %len(folders))
    return folders


def sniff_header(dicom_file):
    '''sniff_header will return an estimated affine, x,y,z dimension from
    the first image (or any) in a series of dicom
    '''
    header = dict()
    ds = dicom.read_file(dicom_file)
    header['xdim'] = ds.Rows
    header['ydim'] = ds.Columns
    header['window_center'] = int(ds.WindowCenter)
    return header
    

def read_series(dicoms,return_nifti=True):
    '''read_series will read in a series of dicoms belonging to a group
    :param dicoms: a list of dicom files to parse, assumed in same series and
    equal size
    :param return_nifti: If True (default) will turn image as Nifti File
    '''
    # Sort the dicoms
    dicoms.sort()

    # Get the size of the image
    params = sniff_header(dicoms[0])
    xdim = params['xdim']
    ydim = params['ydim']   
    windom_center = params['window_center']
 
    bot.debug("First dicom found with dimension %s by %s, using as standard." %(xdim,ydim))

    # Let's get ordering of images based on InstanceNumber
    ordered = dict()
    for d in range(len(dicoms)):    
        ds = dicom.read_file(dicoms[d])
        if ds.Rows == xdim and ds.Columns == ydim:
            ordered[int(ds.InstanceNumber)] = ds.pixel_array

    # Sort by order
    zdim = len(ordered)    
    data = numpy.ndarray((xdim,ydim,zdim))

    # Start at window center, then go back to zero
    index = 0
    for key in sorted(ordered.keys()):
        data[:,:,index] = ordered[key]
        index +=1
    
    if return_nifti == True:
        affine = numpy.diag((1,1,1,0))
        data = nibabel.Nifti1Image(data,affine=affine)
    return data


def dicom2nifti(folders,outdir=None,extension=None):
    '''dicom2nifti will take a list of folders and produce nifti files
    in an output directory. If not defined, they will be output in their
    original directory.
    '''
    if isinstance(folders,dict):
        folders = list(folders.keys())
    
    if not isinstance(folders,list):
        folders = [folders]

    outfiles = []
    for folder in folders:
        lookup = find_dicoms(folder,extension)
        for base,dicomlist in lookup.items():
            nii = read_series(dicomlist)
            if outdir != None:
                outfile = "%s/%s.nii.gz" %(outdir,os.path.basename(base))
            else:
                outfile = "%s/%s.nii.gz" %(base,os.path.basename(base))
            bot.info("Saving %s" %outfile)
            nibabel.save(nii,outfile)
            outfiles.append(outfile)
    return outfiles


if __name__ == '__main__':
    folder = os.getcwd()
    dicoms = find_dicoms(folder)



'''
tasks.py: functions to extract identifiers from dicom images

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


from som.logger import bot
from som.utils import (
    read_json
)

import dateutil.parser
import tempfile

from deid.identifiers import (
    get_timestamp
)

from deid.dicom import (
    get_files,
    get_identifiers as get,
    replace_identifiers as put,
)

from .settings import (
   entity as entity_options,
   item as item_options
)

import os
import sys
here = os.path.dirname(os.path.abspath(__file__))


######################################################################
# MAIN GET FUNCTIONS
######################################################################

def get_identifiers(dicom_files,force=True):
    '''extract and validate identifiers from a dicom image This function 
    uses the deid standard get_identifiers, and formats the data into 
    what is expected for the SOM API request. 
    :param dicom_files: the dicom file(s) to extract from
    :param force: force reading the file (default True)
    '''
    # Enforce application default
    entity_id = entity_options['id_source']
    item_id = item_options['id_source']

    # get_identifiers: returns ids[entity][item] = {"field":"value"}
    ids = get(dicom_files=dicom_files,
              force=force,
              entity_id=entity_id,
              item_id=item_id)


    bot.verbose("Starting preparation of %s entity for SOM API" %(len(ids)))
    entity_cf = entity_options['custom_fields']
    item_cf = item_options["custom_fields"]
    entity_date = entity_options["PatientBirthDate"]
    entity_times = entity_options['id_timestamp']
    item_times = item_options['id_timestamp']
    

    # Now we build a request from the ids
    request = dict()
    for eid,items in ids.items():
        request[eid] = {"id_source":entity_id,
                        "id":eid,
                        "items":[],
                        "custom_fields":{}}

        bot.debug('entity id: %s' %(eid))
        for iid,item in items.items():
            bot.debug('item id: %s' %(iid))

            # Here we generate a timestamp for the entity
            if "id_timestamp" not in request[eid]:
                entity_ts = get_timestamp(item_date=entity_times['date'])
                request[eid]['id_timestamp'] = entity_ts

            # Generate the timestamp for the item
            item_ts = get_timestamp(item_date=item_times['date'],
                                    item_time=item_times['time'])

            new_item = {"id_source": item_id,
                        "id": iid,
                        "custom_fields":{}}

            for header,value in item.items():
            
                # Is it wanted for the entity?
                if header in entity_cf:
                    request[eid]['custom_fields'][header] = value

                # Is it wanted for the item?
                elif header in item_cf:
                    new_item["custom_fields"][header] = value

            request[eid]["items"].append(new_item)
       
    
    # Upwrap the dictionary to return an identifiers objects with a list of all entities
    ids = {"identifiers": [entity for key,entity in request.items()]}
    return ids




def replace_identifiers(response,dicom_files,force=True,deid=None,
                        output_folder=None,overwrite=True):
    '''replace identifiers will replace dicom_files with a response
    from the identifiers API. 
    :param response: the response from the API, or a list of identifiers
    :param dicom_files: the dicom file(s) to extract from
    :param force: force reading the file (default True)
    :param deid: The deid (recipe for replacement) defaults to SOM deid
    :param output_folder: if not defined, uses temp. directory
    :param overwrite: if False, save updated files to temporary directory

    :: note a response looks like this

    [{'id': '14953772',
      'id_source': 'Stanford MRN',
      'items': [{'custom_fields': [{'key': 'ordValue', 'value': '33.1'}],
        'id': 'MCH',
        'id_source': 'Lab Result',
        'jitter': -19,
        'jittered_timestamp': '2010-01-16T11:50:00-0800',
        'suid': '10f6'}],
      'jitter': -19,
      'jittered_timestamp': '1961-07-08T00:00:00-0700',
      'suid': '10f5'}]
    '''

    # Generate ids dictionary for data put (replace_identifiers) function
    ids = dict()

    # Enforce application default
    entity_id = entity_options['id_source']
    item_id = item_options['id_source']

    if deid is None:
        deid = "%s/deid.som" %(here)

    if not os.path.exists(deid):
        bot.warning("Cannot find deid %s, using defaults." %(deid))


    # Format data correctly for deid
    ids = dict()

    for entity in response:

        eid = entity['id']
        bot.debug('entity id: %s' %(eid))
        ids[eid] = dict()

        for item in result['items']:
            iid = item['id']
            bot.debug('item id: %s' %(iid))

            # Custom variables
            ids[eid][iid] = {'item_timestamp': item['jittered_timestamp'],
                             'entity_timestamp': entity['jittered_timestamp'],
                             'entity_id': entity['suid'],
                             'item_id': item['suid'] }


            # All custom fields (likely not used)
            for field in item['custom_fields']:
                ids[eid][iid][field['key']] = field['value'] 

           
    # Do the request to update the files, get them
    updated_files = put(dicom_files,
                        ids=ids,
                        deid=deid,
                        overwrite=overwrite,
                        output_folder=output_folder,
                        entity_id=entity_id,
                        item_id=item_id,
                        force=force)
                        
    return updated_files
